#
# Implemets redirections for success or error in controllers.
#
module IncidentSetters
  extend ActiveSupport::Concern

  include IncidentsHelper


  def set_elements
    search_params = {}
    @elements = @incident.get_data(search_params)

    if !@incident.list_linked_incidents.empty?
      @incident.list_linked_incidents.each do |linked_incident|
        @elements.concat(linked_incident.get_data(search_params))
      end
    end
  end

  def load_indicators(elements)
    @indicators = Hash.new(0)

    grouped_elements = elements.group_by { |event| [event[:msg], event[:incident_source_name], event[:sig_id], event[:classification]] }

    @indicators = grouped_elements.map do |(msg, incident_source_name, sig_id, classification), entries|
      indicator = {
        incident_source_name: incident_source_name,
        msg: msg,
        classification: classification,
        count: entries.size,
        sig_id: sig_id
      }
      indicator
    end.compact

    @indicators = @indicators.sort_by { |entry| -entry[:count] }
  end

  def load_actions_and_worklog_data
    @actions = gather_all_actions(@incident)
    @worklog_data = sort_worklog_data(gather_all_worklog_data(@incident), :desc)
    @actions_taken_data = fetch_actions_taken_data(@incident)
    @actions_taken_data = sort_worklog_data(@actions_taken_data, :desc)
  end

  def calculate_nodes
    @incident_nodes = []
    @timestamps = []

    process_assets(@elements)
    process_observables(@elements)
    process_timestamps(@elements)

    @assets = group_and_count(@incident_nodes, :asset)
    @observables = group_and_count(@incident_nodes, :observable)
    @observables_by_type = @observables.group_by { |obs| obs[:key][:type] }
  end

  def gather_all_actions(incident)
    response_actions = incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions)
    all_actions = (response_actions + response_actions).sort_by(&:created_at)
    all_actions
  end

  def gather_all_worklog_data(incident)
    incident_changes = gather_incident_audit_worklog_data(incident)
    notes = gather_response_action_notes_worklog_data(incident) + gather_notes_worklog_data(incident)
    executions = gather_response_action_executions_worklog_data(incident)
    all_worklog_data = (incident_changes + notes + executions)
    all_worklog_data
  end

  def gather_incident_audit_worklog_data(incident)
    incident.audits.map do |audit|
      formatted_content = audit.comment

      if audit.audited_changes['incident_status_id']
        # Obtener los valores antiguos y nuevos del estado
        old_status_id, new_status_id = audit.audited_changes['incident_status_id']

        # Buscar los nombres de los estados antiguos y nuevos
        old_status = IncidentStatus.find_by(id: old_status_id).try(:name) ? IncidentStatus.find_by(id: old_status_id).name : old_status_id
        new_status = IncidentStatus.find_by(id: new_status_id).try(:name) ? IncidentStatus.find_by(id: new_status_id).name : new_status_id

        # Reemplazar el texto en el comentario
        # TODO: refactor this to cover all languages..
        # English
        formatted_content = formatted_content.gsub(/incident_status_id from '(\d+)' to '(\d+)'/) do
          "status from '#{old_status}' to '#{new_status}'"
        end
        # Spanish
        formatted_content = formatted_content.gsub(/incident_status_id desde '(\d+)' a '(\d+)'/) do
          "estado desde '#{old_status}' a '#{new_status}'"
        end
      end

      {
        action: audit.action,
        user: audit.user,
        created_at: audit.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: formatted_content
      }
    end
  end

  def gather_response_action_notes_worklog_data(incident)
    return incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map(&:notes).map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        source: "[RESPONSE TASK] #{note.response_action.task.name}",
        content: note.content,
        type: "note",
        note_id: note.id,
        incident_id: incident.id
      }
    end
  end

  def gather_notes_worklog_data(incident)
    return incident.notes.map do |note|
      {
        action: "Create",
        user: note.user,
        created_at: note.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
        content: note.content,
        type: "note",
        note_id: note.id,
        incident_id: incident.id
      }
    end
  end

  def gather_response_action_executions_worklog_data(incident)
    incident.playbook.phases.flat_map(&:tasks).flat_map(&:response_actions).flat_map do |response_action|
      if response_action.notes.any?
        []
      else
        stored_job = StoredDelayedJob.where("options LIKE ?", "%response_action_id: #{response_action.id}%").first
        stored_job_id = stored_job.present? ? stored_job.id : nil
        [{
          action: "Executed Action",
          user: response_action.user,
          created_at: response_action.created_at.strftime("%B %d, %Y %I:%M:%S %p"),
          source: "[RESPONSE TASK] #{response_action.task.name}",
          stored_job_id: stored_job_id
        }]
      end
    end.compact
  end

  def sort_worklog_data(worklog_data, order)
    date_format = "%B %d, %Y %I:%M:%S %p"

    converted_data = worklog_data.map do |item|
      item[:created_at_datetime] = DateTime.strptime(item[:created_at], date_format)
      item
    end

    sorted_data = converted_data.sort_by { |item| item[:created_at_datetime] }

    sorted_data.reverse! if order == :desc

    sorted_data.each { |item| item.delete(:created_at_datetime) }

    sorted_data
  end

  def fetch_actions_taken_data(incident)
    data = []
    notes = gather_response_action_notes_worklog_data(incident)
    notes.each do |note|
      data.push(format_actions_taken_note(note))
    end

    executions = gather_response_action_executions_worklog_data(incident)

    executions.each do |execution|
      data.push(format_actions_taken_execution(execution))
    end

    data
  end

  def process_assets(elements)
    elements.each do |element|
      next if element[:assets].blank?
      timestamp = element[:timestamp]

      element[:assets].each do |asset_element|
        existing_node = find_existing_node(asset_element[:name], :asset)

        if existing_node
          add_timestamp(existing_node, timestamp)
        else
          @incident_nodes.push({
            name: asset_element[:name],
            type: asset_element[:type],
            value: asset_element[:mac_address],
            id: asset_element[:id],
            image: ActionController::Base.helpers.get_asset_image(asset_element[:type]),
            asset: true,
            timestamps: [timestamp]
          })
        end
      end
    end
  end

  def process_observables(elements)
    elements.each do |element|
      next if element[:observables].blank?
      timestamp = element[:timestamp]

      element[:observables].each do |observable_element|
        observable_element.each do |key, value|
          existing_node = find_existing_node(key, :observable)

          if existing_node
            add_timestamp(existing_node, timestamp)
          else
            @incident_nodes.push({
              name: key,
              type: value,
              image: ActionController::Base.helpers.get_observable_image(value, nil),
              observable: true,
              timestamps: [timestamp]
            })
          end
        end
      end
    end
  end

  def find_existing_node(name, type_key)
    @incident_nodes.find { |node| node[:name] == name && node[type_key] }
  end

  def add_timestamp(node, timestamp)
    node[:timestamps] << timestamp unless node[:timestamps].include?(timestamp)
  end

  def process_timestamps(elements)
    elements.each do |element|
      timestamp = element[:timestamp]
      @timestamps << timestamp
    end

    @timestamps = @timestamps.sort
  end

  def group_and_count(nodes, type)
    grouped_nodes = nodes.select { |node| node[type] == true }.group_by { |e| e }
    grouped_nodes.map { |key, value| { key: key, count: key[:timestamps].size } }.sort_by { |e| -e[:count] }
  end

  # Set @sensor_names for all actions
  def set_sensor_names
    @sensor_names = current_user.domains_with_permission.map { |d| [d.id, d.path_name] }.to_h
  end

end
