#######################################################################
# Copyright (c) 2025 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class InventoryDevicesController < ApplicationController
  helper Redborder::ModulesHelper

  include ApplicationHelper
  include InventoryDevicesHelper

  # Storage of the filters
  include Redborder::Filters::Storage::Session

  ## ======================================================================
  # => Before actions and filters
  ## ======================================================================
  before_action :set_inventory_device, only: [:destroy, :show, :edit, :update]
  before_action :set_sensor_names, only: [:index, :search, :show, :edit, :create, :statistics]
  before_action :set_criticalities, only: [:index, :search, :show, :edit]
  before_action :set_range, only: [:show]

  layout false, only: [:new]

  #
  # => Render all assets in a table
  #
  def index
    @page = params[:page] || 1

    # Get only the assets where the sensor_id is among those allowed
    @inventory_devices = InventoryDevice.where(sensor_id: @sensor_names.keys).order(last_seen: :desc)

    @total_inventory_devices_count = @inventory_devices.count

    # Apply cached filters by reusing search method if it is present
    @cached_filters = InventoryDevice.read_filters(current_user.id)
    if @cached_filters.present?
      params[:search] = @cached_filters[:search]
      params[:name] = @cached_filters[:name]
      @inventory_devices = search
    end
    
    # To show statistics on index view
    # incident_counters

    # Paginate results
    @inventory_devices = @inventory_devices.paginate(page: @page, per_page: current_user.per_page_count)
  end

  def new
    @inventory_device = InventoryDevice.new
  end

  def create
    @inventory_device = InventoryDevice.new(inventory_device_params)
    @inventory_device.sources = ["User"]

    if @inventory_device.save
      @page = params[:page] || 1

      @inventory_devices = InventoryDevice.where(sensor_id: @sensor_names.keys).order(last_seen: :desc)

      @total_inventory_devices_count = @inventory_devices.count

      # Apply cached filters by reusing search method if it is present
      @cached_filters = InventoryDevice.read_filters(current_user.id)
      if @cached_filters.present?
        params[:search] = @cached_filters[:search]
        params[:name] = @cached_filters[:name]
        @inventory_devices = search
      end

      @inventory_devices = @inventory_devices.paginate(page: @page, per_page: current_user.per_page_count)
      flash.now[:notice] = "Device was created successfully"
    else
      flash[:alert] = "Cannot create device"
    end
  end

  def show
    @top_applications = top_applications_query
    @top_lan_ports = top_lan_ports_query
    @top_wan_ports = top_wan_ports_query

    @incidents = @inventory_device.incidents
      .where(sensor_id: @sensor_names.keys)
      .where(updated_at: @start_time..@end_time)
      .order(created_at: :desc)
      .limit(5)
  end

  def statistics
    created_at_condition = if current_user.time_machine_now
      time_machine = Time.zone.at(current_user.time_machine_now)
      ["created_at <= ?", time_machine]
    else
      nil
    end

    # Incindents by criticality
    @assets_counts_by_criticality = Criticality.all.map do |criticality|
      {
        name: criticality.name.capitalize,
        events: InventoryDevice.where(
          criticality_id: criticality.id,
          sensor_id: @sensor_names.keys
        ).where(created_at_condition).count
      }
    end.select { |priority_data| priority_data[:events] > 0 }.sort_by { |asset_count| -asset_count[:events] }

    # Incindents by type
    @assets_counts_by_type = InventoryDeviceTypeObject.all.map do |type|
      {
        name: type.name.capitalize,
        events: InventoryDevice.where(
          inventory_device_type_object_id: type.id,
          sensor_id: @sensor_names.keys
        ).where(created_at_condition).count
      }
    end.select { |priority_data| priority_data[:events] > 0 }.sort_by { |asset_count| -asset_count[:events] }

    # Incindents by OS
    @assets_counts_by_os = OperatingSystemObject.all.map do |os|
      {
        name: os.product.capitalize + ' ' + os.version.capitalize,
        events: InventoryDevice.where(
          operating_system_object_id: os.id,
          sensor_id: @sensor_names.keys
        ).where(created_at_condition).count
      }
    end.select { |priority_data| priority_data[:events] > 0 }.sort_by { |asset_count| -asset_count[:events] }

    # Incindents by source
    @assets_counts_by_source = InventoryDevice.where(sensor_id: @sensor_names.keys)
        .where(created_at_condition)
        .group(:sources)
        .count
        .map do |source, count|
          {
            name: source.join('/'),
            events: count
          }
    end

    # Incindents by hardware
    @assets_counts_by_hardware = HardwareObject.all.map do |hardware|
      {
        name: hardware.product,
        events: InventoryDevice.where(
          hardware_object_id: hardware.id,
          sensor_id: @sensor_names.keys
        ).where(created_at_condition).count
      }
    end.select { |priority_data| priority_data[:events] > 0 }.sort_by { |asset_count| -asset_count[:events] }

    # Incindents by domain
    @assets_counts_by_domain = InventoryDevice.where(sensor_id: @sensor_names.keys)
        .where(created_at_condition)
        .group(:sensor_id)
        .count
        .map do |sensor_id, count|
          {
            name: @sensor_names[sensor_id] == 'root' ? '/' : @sensor_names[sensor_id],
            events: count
          }
    end
  end

  def force_update
    DelayedJob.force_update_inventory_devices
    redirect_to inventory_devices_path
  end

  def check_update_inventory_devices_job_job_status
    running = update_inventory_devices_job_executing?
    render json: { running: running }
  end

  #
  # Search on assets
  #
  def search(paginate = true, widget_mode = false)
    @page = params[:page] || 1
    @inventory_devices = InventoryDevice.where(sensor_id: @sensor_names.keys)
    
    # If params[:search] or params[:name] exists, save the current search filters to Memcached
    unless widget_mode
      if params[:search].present? || params[:name].present?
        InventoryDevice.create_filters(current_user.id, params[:search], params[:name])
      else
        # Retrieve filters from Memcached if no search params are provided
        cached_filters = InventoryDevice.read_filters(current_user.id)
        if cached_filters.present?
          params[:search] = cached_filters[:search]
          params[:name] = cached_filters[:name]
        end
      end
    end

    # Filter by name
    if params[:name].present?
      search_term = "%#{params[:name]}%"
      @inventory_devices = @inventory_devices.left_joins(:mac_objects).where(
        "inventory_devices.name ILIKE :term OR array_to_string(inventory_devices.ips, ',') ILIKE :term OR inventory_devices.id::text ILIKE :term OR redborder_objects.value ILIKE :term OR redborder_objects.vendor ILIKE :term OR inventory_devices.comments ILIKE :term",
        term: search_term
      ).distinct
    end
    if params[:search].present?
      # Filter by criticality
      if params[:search][:criticality].present?
        @inventory_devices = @inventory_devices.where(criticality_id: params[:search][:criticality])
      end

      # Filter by domain
      if params[:search][:domain].present?
        @inventory_devices = @inventory_devices.where(sensor_id: params[:search][:domain])
      end

       # Filter by inventory device type object
       if params[:search][:inventory_device_type_object].present?
        @inventory_devices = @inventory_devices.where(inventory_device_type_object_id: params[:search][:inventory_device_type_object])
      end

      if params[:search][:is_locked].present?
        @inventory_devices = @inventory_devices.where(locked: params[:search][:is_locked])
      end

      # Filter by minimum criticality
      if params[:search][:minimum_criticality].present?
        selected_criticality_id = params[:search][:minimum_criticality].to_i
        selected_criticality_score = Criticality.where(id: selected_criticality_id).pluck(:score)
        higher_criticality_ids = Criticality.where('score > ?', selected_criticality_score).pluck(:id)
        @inventory_devices = @inventory_devices.where(criticality_id: higher_criticality_ids)
      end

      # Filter by source
      if params[:search][:sources].present?
        value_to_search = nil
        if params[:search][:sources].kind_of?(Array)
          value_to_search = params[:search][:sources].map(&:downcase)
        elsif params[:search][:source].kind_of?(String)
          value_to_search = [params[:search][:sources]].map(&:downcase)
        end
        @inventory_devices = @inventory_devices.where('EXISTS (SELECT 1 FROM unnest(sources) s WHERE LOWER(s) IN (?))', value_to_search)
      end

      # Exclude assets by name (not like)
      if params[:search][:exclude].present?
        exclude_query = params[:search][:exclude].split(',').map { |term| "inventory_devices.name NOT ILIKE ?" }.join(' AND ')
        exclude_values = params[:search][:exclude].split(',').map { |term| "%#{term.strip}%" }
        @inventory_devices = @inventory_devices.where(exclude_query, *exclude_values)
      end

      # Handle sorting
      order_by = params[:search][:order_by]
      direction = params[:search][:order_direction] == 'desc' ? 'desc' : 'asc'
    end

    if params[:search][:order_by].present?
      # For other fields, apply safe sorting
      @inventory_devices = @inventory_devices.order(InventoryDevice.arel_table[order_by].send(direction))
    else
      # Fallback to a default order if the column is not allowed
      @inventory_devices = @inventory_devices.order("inventory_devices.created_at DESC")
    end

    if paginate
      # Paginate results
      @inventory_devices = @inventory_devices.paginate(page: @page, per_page: current_user.per_page_count)
    else
      @inventory_devices
    end
  end

  def reset_filters
    InventoryDevice.clear_filters(current_user.id)
    redirect_to inventory_devices_path
  end

  #
  # => Destroy an asset
  #
  def destroy
    @inventory_device.destroy
    respond_to do |format|
      format.html { redirect_to inventory_devices_path }
      format.js
      format.json { head :no_content }
    end
  end

  def bulk_delete
    page = params[:page] || 1
    inventory_device_ids = (params[:inventory_device_ids] || []).reject(&:blank?).map(&:to_i)

    if inventory_device_ids.any?
      InventoryDevice.where(id: inventory_device_ids).destroy_all
      flash[:notice] = "Assets were deleted successfully"
    else
      flash[:alert] = "No assets were deleted"
    end

    redirect_to inventory_devices_path(page: page)
  end

  def bulk_update_criticality
    page = params[:page] || 1
    inventory_device_ids = (params[:inventory_device_ids] || []).reject(&:blank?).map(&:to_i)
    criticality = Criticality.find_by(name: params[:criticality])
    inventory_devices = InventoryDevice.where(id: inventory_device_ids)

    if criticality && inventory_devices.update_all(criticality_id: criticality.id)
      flash[:notice] = "Assets criticality updated successfuly"
    else
      flash[:alert] = "Error updating assets criticality"
    end

    redirect_to inventory_devices_path(page: page)
  end

  def edit
  end

  def update
    if @inventory_device.update(inventory_device_params)
      redirect_to @inventory_device, notice: "Asset " + I18n.t('controllers.successfully_updated')
    else
      flash[:alert] = I18n.t('controllers.error_updating')
      render :edit
    end
  end

  def inventory_devices_to_csv_format
    columns = InventoryDevice.column_names + ['mac_objects']
    CSV.generate(col_sep: ";") do |csv|
      if current_user.admin?
        csv << columns
        inventory_devices = InventoryDevice.where(sensor_id: current_user.domains_with_permission)
        inventory_devices.order(created_at: :desc).each do |device|
          csv << columns.map { |col|
            if col == 'mac_objects'
              mac_objects = device.send(col)
              mac_objects.map { |mac| mac.value}
            else
              device.send(col)
            end
          }
        end
      end
    end
  end

  def export
    filename = "#{Time.now.strftime("%Y%m%d%H%M%S")}-Assets.csv"
    content = inventory_devices_to_csv_format
    send_data content, filename: filename, type: 'text/csv'
  end

  def import
    if current_user.super_admin?
      render layout: false
    else
      error_500('')
    end
  end

  def import_file
    content_file = params[:inventory_device][:csv_file]
    
    unless content_file.content_type == "text/csv"
      redirect_to inventory_devices_path, alert: t("inventory_devices.import.csv_error")
      return
    end
  
    begin
      created_count, updated_count = import_from_csv(content_file)
      flash[:notice] = t("inventory_devices.import.import_success")
    rescue => e
      flash[:alert] = t("inventory_devices.import.csv_error", error: e.message)
    end
    
    redirect_to inventory_devices_path
  end

  def import_from_csv(content_file)
    created_count = 0
    updated_count = 0
    csv_data = CSV.read(content_file.path, col_sep: ';', headers: true, encoding: 'UTF-8', liberal_parsing: true)
    columns = csv_data.headers.map(&:to_sym)

    CSV.foreach(content_file.path, col_sep: ';', headers: false, encoding: 'UTF-8', liberal_parsing: true) do |row|
      next if row.empty? || row[0].nil?

      attrs = {}
      columns.each_with_index do |col, index|
        attrs[col] = row[index].try(:strip)
      end

      attrs[:created_at] = Time.now

      next unless attrs[:id].to_i != 0

      # Parse all the columns that are arrays
      attrs[:sources] = JSON.parse attrs[:sources]
      attrs[:ips] = JSON.parse attrs[:ips]

      # Parse the macs string to convert it to an array, then check if the mac exists, if not, create it
      mac_addresses = JSON.parse attrs[:mac_objects]
      attrs[:mac_objects] = mac_addresses.map { |mac| 
        mac_object = MacObject.new(
          value: mac,
          name: (attrs[:name] || mac),
          vendor: nil,
          user_id: nil,
          created_at: Time.current,
          updated_at: Time.current,
          sensor_id: attrs[:sensor_id]
        )
        # Save the new object to the database
        mac_object.save!
        mac_object
      }

      object = InventoryDevice.find_or_initialize_by(id: attrs[:id])

      if object.new_record?
        object.assign_attributes(attrs)
        object.save
        created_count += 1
        # This is necessary because the id has been passed manually to create the new record, therefore 
        # the sequence of ids must be updated with the id of the record created.
        ActiveRecord::Base.connection.reset_pk_sequence!('inventory_devices')
      else
        attrs.delete(:created_at)
        object.update(attrs)
        updated_count += 1
      end
    end
  end
  
  private

  def set_criticalities
    @criticalities = Criticality.all.order(score: :desc)
  end

  # Set @sensor_names for all actions
  def set_sensor_names
    @sensor_names = current_user.domains_with_permission.map { |d| [d.id, d.path_name] }.to_h
  end

  #
  # => Init the asset if the user has permissions
  #
  def set_inventory_device
    @inventory_device = InventoryDevice.find(params[:id])
  end

  def inventory_device_params
    params.require(:inventory_device).permit(
      :criticality_id, :name, { sources: [] }, :sensor_id, :locked, :comments, 
      :inventory_device_type_object_id, :hardware_object_id, :operating_system_object_id, 
      mac_objects_attributes: [:id, :value, :name, :sensor_id, :_destroy], ips: []
    )
  end

  def top_applications_query
    return [] unless @inventory_device.mac_objects.any?

    return [] unless Redborder::Module.registered_modules.key? :flow

    begin
      source = Redborder::DruidSource.get(source: 'flow')

      filter = Redborder::Filter.new(@source, {})
      @inventory_device.mac_objects.map(&:value).each do |mac|
        filter.add('client_mac', mac)
      end

      @granularity = (params[:granularity] || :minute)

      @source.topN(
        query_type: :group_by,
        start_time: @start_time,
        end_time: @end_time,
        range: @range,
        granularity: :all,
        limit: 10,
        group_by: source::COLUMNS[:application_id_name][:columns].keys,
        aggregation: :bytes,
        search: filter.get
      )
    rescue => e
      Rails.logger.error "ERROR: #{e.message}"
      return []
    end
  end

  def top_lan_ports_query
    return [] unless @inventory_device.mac_objects.any?

    return [] unless Redborder::Module.registered_modules.key? :flow

    begin
      source = Redborder::DruidSource.get(source: 'flow')

      filter = Redborder::Filter.new(@source, {})
      @inventory_device.mac_objects.map(&:value).each do |mac|
        filter.add('client_mac', mac)
      end

      @granularity = (params[:granularity] || :minute)

      @source.topN(
        query_type: :group_by,
        start_time: @start_time,
        end_time: @end_time,
        range: @range,
        granularity: :all,
        limit: 10,
        group_by: source::COLUMNS[:lan_l4_port][:columns].keys,
        aggregation: :bytes,
        search: filter.get
      )
    rescue => e
      Rails.logger.error "ERROR: #{e.message}"
      return []
    end
  end

  def top_wan_ports_query
    return [] unless @inventory_device.mac_objects.any?

    return [] unless Redborder::Module.registered_modules.key? :flow

    begin
      source = Redborder::DruidSource.get(source: 'flow')

      filter = Redborder::Filter.new(@source, {})
      @inventory_device.mac_objects.map(&:value).each do |mac|
        filter.add('client_mac', mac)
      end

      @granularity = (params[:granularity] || :minute)

      @source.topN(
        query_type: :group_by,
        start_time: @start_time,
        end_time: @end_time,
        range: @range,
        granularity: :all,
        limit: 10,
        group_by: source::COLUMNS[:wan_l4_port][:columns].keys,
        aggregation: :bytes,
        search: filter.get
      )
    rescue => e
      Rails.logger.error "ERROR: #{e.message}"
      return []
    end
  end

  def set_range
    return unless Redborder::Module.registered_modules.key? :flow

    @source = Redborder::DruidSource.get(source: 'flow') unless @source
    
    # Recover filters from session (usefull for custom range on time selection)
    @filter = Redborder::Filter.new(@source, load_filters)
    
    # Keep only start_time and end_time filters
    @filter.get.each_key do |column|
      @filter.remove(column) unless [:start_time, :end_time].include?(column)
    end
    save_filters(@filter.filters) 

    if params[:range] && params[:range] != 'custom'
      @filter.remove(:start_time)
      @filter.remove(:end_time)
      save_filters(@filter.filters)
      @range = params[:range]
    elsif params[:range]
      if params[:range] == 'custom' && !@filter.exist_any?(:start_time) &&
         !@filter.exist_any?(:end_time)
        @range = 'last_1_hour'
      else
        @range = params[:range]
      end
    else
      @range = (@global_filters || {}).fetch(:range, 'last_1_hour')
    end

    end_time = start_time = Time.now

    in_time_machine do
      end_time = Time.now
      start_time = end_time - 1.hour
      if @filter.exist_any?(:start_time) && @filter.exist_any?(:end_time)
        @range = 'custom'

        start_time = Time.at(@filter.get(:start_time).keys.first)

        end_time = Time.at(@filter.get(:end_time).keys.first)
      end

      @end_time = Time.now.change(sec: 0)
    end

    case @range.to_sym
    when :live
      @start_time = @end_time - 3.minutes
    when :last_5_minute
      @start_time = @end_time - 5.minutes
    when :last_15_minute
      @start_time = @end_time - 15.minutes
    when :last_30_minute
      @start_time = @end_time - 30.minutes
    when :last_1_hour
      @start_time = @end_time - 1.hour
    when :last_2_hour
      @start_time = @end_time - 2.hour
    when :last_3_hour
      @start_time = @end_time - 3.hour
    when :last_12_hour
      @start_time = (@end_time - 12.hour).beginning_of_hour
    when :last_24
      @start_time = (@end_time - 24.hour).beginning_of_hour
    when :last_week
      @start_time = (@end_time - 1.week).beginning_of_hour
    when :custom
      @end_time   = end_time
      @start_time = start_time
    else
      @start_time = @end_time - 1.hour
    end

    @start_time_for_slider = @start_time
    @end_time_for_slider = @end_time

    granularities = Redborder::Rbdruid
                    .granularity_list(start_time: @start_time,
                                      end_time: @end_time)
    granularity = if params[:granularity].present?
                    params[:granularity]
                  else
                    granularities.first
                  end

    # Multiple return lines as granularity.to_sym should not be executed
    # if params[:ganularity] is not a valid one
    if params[:w]
      ## For all widgets
      return if Redborder::Rbdruid::GRANULARITY_TYPE.keys.map(&:to_s).exclude?(granularity)
    end

    return if granularity.to_sym.eql?(:all)

    step = Redborder::Rbdruid::GRANULARITY_TYPE[granularity.to_sym][:step]
    roundto = step / 1000
    @start_time = Time.at((@start_time.to_i / roundto) * roundto)
  end

end