#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

module AlarmsHelper
  def int_to_hours num
    ret = num.to_i / 60
    if ret > 59
      (ret / 60).to_s + " hours"
    else
      ret.to_s + " minutes"
    end
  end

  def get_number_from_severity

    unless @alarm.severity.nil?

        case @alarm.severity.name
          when 'very low'
            # Very low
            return 0
          when 'low'
            # Low
            return 1
          when 'medium'
            # Medium
            return 2
          when 'high'
            # High
            return 3
          when 'very high'
            # Very High
            return 4
          when 'critical'
            # Critical
            return 5
          else
            # Very low by default
            return 0
          end
    end

    # By default, very low
    return 0

  end

  def alarms_dropdown_options
    html =
      content_tag(:li) do
        link_to t("import_csv"),
                import_alarms_path,
                class: "import_alarms",
                "data-toggle" => "modal",
                "data-target" => "#default_modal",
                "data-backdrop" => "static"
      end
  
    html +=
      content_tag(:li) do
        link_to t("export_csv"),
                export_alarms_path,
                class: "export_alarms"
      end
  
    # Devuelve la cadena HTML segura
    html.html_safe
  end


  def self.process_query_data(query_data, alarm)
    product_type = alarm.product_type
    case product_type
    when 'ap_state', 'wireless'
      process_ap_state(query_data, alarm.aggregation)
    when 'flow'
      process_flow(query_data, alarm.aggregation)
    when 'ips'
      process_ips(query_data, alarm.aggregation)
    when 'monitor'
      process_monitor(query_data)
    when 'vault'
      process_vault(query_data, alarm.aggregation)
    when 'wireless'
      process_wireless(query_data)
    else
      default_processing(query_data)
    end
  end

  def self.normalize_query_data(query_data)
    if query_data.is_a?(Hash) && query_data.key?(:more_info)
      query_data[:more_info]
    elsif query_data.is_a?(String)
      begin
        parsed = JSON.parse(query_data, symbolize_names: true)
        parsed.is_a?(Array) ? parsed : [parsed]
      rescue JSON::ParserError
        []
      end
    else
      query_data
    end
  end
  
  def self.process_ap_state(query_data, aggregation)
    data = normalize_query_data(query_data)
  
    unless data.is_a?(Array) && data.all? { |ap_state| ap_state.respond_to?(:wireless_station) }
      Rails.logger.error "Datos inválidos en process_ap_state: #{data.inspect}"
      return []
    end
  
    grouped = data.group_by { |ap_state| ap_state.wireless_station }
  
    grouped.map do |station, ap_states|
      aggregated_value = ap_states.sum do |ap|
        if ap.is_a?(Hash)
          ap[aggregation.to_sym].to_f
        else
          ap.respond_to?(aggregation) ? ap.send(aggregation).to_f : 0
        end
      end
  
      first_ap = ap_states.first
  
      result = {
        wireless_station: station,
        sensor:           first_ap.sensor_name,
        aggregation:      aggregated_value,
        timestamp:        first_ap.timestamp
      }

      result[:wireless_station_name] = first_ap.wireless_station_name if first_ap.wireless_station_name
      result
    end
  end
  
  def self.process_vault(query_data, aggregation)
    data = normalize_query_data(query_data)
  
    unless data.is_a?(Array) && data.all? { |vault|
             vault.respond_to?(:source) &&
             vault.respond_to?(:target)
           }
      Rails.logger.error "Datos inválidos en process_vault: #{data.inspect}"
      return []
    end
  
    grouped = data.group_by do |vault|
      source = vault.source
      target = vault.target
      [source, target]
    end
  
    grouped.map do |(source, target), vaults|
      aggregated_value = vaults.sum do |vault|
        if vault.is_a?(Hash)
          vault[aggregation.to_sym].to_f
        else
          vault.respond_to?(aggregation) ? vault.send(aggregation).to_f : 0
        end
      end
  
      vault = vaults.first
  
      {
        source:      source,
        target:      target,
        sensor:      vault.sensor_name,
        aggregation: aggregated_value,
        timestamp:   vault.timestamp,
        hostname:    vault.hostname
      }
    end
  end
  
  def self.process_ips(query_data, aggregation)
    data = normalize_query_data(query_data)
  
    unless data.is_a?(Array) && data.all? { |event|
             (event.respond_to?(:src) || event.respond_to?(:src_name)) &&
             (event.respond_to?(:dst) || event.respond_to?(:dst_name))
           }
      Rails.logger.error "Datos inválidos en process_ips: #{data.inspect}"
      return []
    end
  
    grouped = data.group_by do |event|
      source = event.respond_to?(:src) ? event.src : event.src_name
      target = event.respond_to?(:dst) ? event.dst : event.dst_name
      [source, target]
    end
  
    grouped.map do |(source, target), events|
      aggregated_value = events.sum do |event|
        if event.is_a?(Hash)
          event[aggregation.to_sym].to_f
        else
          event.respond_to?(aggregation) ? event.send(aggregation).to_f : 0
        end
      end
  
      event = events.first
  
      {
        source:    source,
        target:    target,
        sensor:    event.sensor_name,
        aggregation: aggregated_value,
        timestamp: event.timestamp,
        msg:       event.msg
      }
    end
  end
  
  def self.process_monitor(query_data)
    data = query_data.is_a?(Hash) && query_data.key?(:more_info) ? query_data[:more_info] : query_data
  
    if data.is_a?(Array) && data.all? { |monitor| monitor.respond_to?(:sum_value) && monitor.respond_to?(:monitors) }
      data.map do |monitor|
        {
          value: monitor.monitors,
          sum: monitor.sum_value,
          min: monitor.min,
          max: monitor.max,
          sensor_name: monitor.sensor_name,
          monitor_sum: monitor.respond_to?(:monitor_sum) ? monitor.monitor_sum : monitor.instance_variable_get(:@monitor_sum),
          timestamp: monitor.timestamp
        }
      end
    else
      Rails.logger.warn "Datos inválidos para Monitor: #{data.inspect}"
      []
    end
  end

  def self.process_flow(query_data, aggregation)
    data = normalize_query_data(query_data)
  
    unless data.is_a?(Array) && data.all? { |traffic|
             (traffic.respond_to?(:lan_ip) || traffic.respond_to?(:lan_ip_name)) &&
             (traffic.respond_to?(:wan_ip) || traffic.respond_to?(:wan_ip_name))
           }
      Rails.logger.error "Datos inválidos en process_flow: #{data.inspect}"
      return []
    end
  
    grouped = data.group_by do |traffic|
      lan = traffic.respond_to?(:lan_ip) ? traffic.lan_ip : traffic.lan_ip_name
      wan = traffic.respond_to?(:wan_ip) ? traffic.wan_ip : traffic.wan_ip_name
      [lan, wan]
    end
  
    grouped.map do |(lan, wan), traffics|
      aggregated_value = traffics.sum do |traffic|
        if traffic.is_a?(Hash)
          traffic[aggregation.to_sym].to_f
        else
          traffic.respond_to?(aggregation) ? traffic.send(aggregation).to_f : 0
        end
      end
  
      traffic = traffics.first
  
      {
        lan_ip:    lan,
        wan_ip:    wan,
        sensor:    traffic.sensor_name,
        aggregation: aggregated_value,
        timestamp: traffic.timestamp
      }
    end
  end

  def self.query_data(data_source, start_time, end_time, filter, alarm)
    if alarm.product_type == "monitor"
      data_source.where(
        query_type: :group_by,
        range: :custom,
        start_time: start_time,
        end_time: end_time,
        granularity: alarm.outliers_enabled ? alarm.outliers_granularity.downcase.to_sym : :all,
        search: filter.get,
        monitor_tab: alarm.monitor_tab,
        dimension: dimensions_for(alarm),
        group_by: ["sensor_name"],
        aggregation: alarm.aggregation,
        user: User.first,
        outliers: alarm.outliers_enabled
      )
    else 
      data_source.where(
      query_type: :group_by,
      range: :custom,
      start_time: start_time,
      end_time: end_time,
      granularity: alarm.outliers_enabled ? alarm.outliers_granularity.downcase.to_sym : :all,
      search: filter.get,
      group_by: dimensions_for(alarm.product_type),
      aggregation: alarm.aggregation,
      user: User.first,
      outliers: alarm.outliers_enabled
    )
    end
  end

  def self.extract_event_info_by_product_type(more_info, product_type)
    case product_type
    when 'ips'
      extract_event_info(more_info)
    when 'flow'
      extract_flow_info(more_info)
    when 'vault'
      extract_vault_info(more_info)
    when 'monitor'
      extract_monitor_info(more_info)
    when 'ap_state', 'wireless'
      extract_ap_state_info(more_info)
    else
      nil
    end
  end

  def self.event_info_fields(event_info, product_type)
    case product_type
    when 'monitor'
      {
        "monitor-max" => event_info[:max],
        "monitor-min" => event_info[:min],
        "monitor-value" => event_info[:value],
        "sensor_name" => event_info[:sensor_name],
      }
    when 'flow'
      {
        "traffic-lan-ip" => event_info[:lan_ip],
        "traffic-wan-ip" => event_info[:wan_ip],
        "aggregation" => event_info[:aggregation],
        "sensor_name" => event_info[:sensor]
      }
    when 'vault'
      {
        "source"      => event_info[:source],
        "target"      => event_info[:target],
        "sensor_name" => event_info[:sensor],
        "aggregation" => event_info[:aggregation],
        "hostname" => event_info[:hostname]
      }
    when 'ap_state', 'wireless'
      {
        "wireless-station"      => event_info[:wireless_station],
        "sensor_name"      => event_info[:sensor],
        "aggregation" => event_info[:aggregation],
      }
    when 'ips'
      {
        "intrusion-source"      => event_info[:source],
        "intrusion-target"      => event_info[:target],
        "sensor_name" => event_info[:sensor],
        "intrusion-msg" => event_info[:msg],
        "aggregation" => event_info[:aggregation]
      }
    else 
      {
        "source"      => event_info[:source],
        "target"      => event_info[:target],
        "sensor_name" => event_info[:sensor],
        "aggregation" => event_info[:aggregation],
        "hostname" => event_info[:hostname]
      }
    end
  end

  def self.extract_aggregation_info(more_info, product_type)
    event_info = extract_event_info_by_product_type(more_info, product_type)

    return nil unless event_info
  
    aggregation_key = (product_type == 'monitor') ? :sum : :aggregation
    aggregation_value = event_info[aggregation_key]
    return nil unless aggregation_value
  
    entities = case product_type
               when 'flow'
                 "the IPs #{event_info[:lan_ip]} and #{event_info[:wan_ip]}"
               when 'ips', 'vault'
                 # Se incluyen solo si ambos están presentes y no vacíos
                 if event_info[:source].present? && event_info[:target].present?
                   "the addresses #{event_info[:source]} and #{event_info[:target]}"
                 end
               when 'monitor'
                 "the sensor #{event_info[:sensor_name]}"
               when 'ap_state', 'wireless'
                 station = event_info[:wireless_station]
                 station_name = event_info[:wireless_station_name]
                 station_name.present? ? "the wireless station #{station} (#{station_name})" : "the wireless station #{station}"
               end
  
    { entities: entities, aggregation_value: aggregation_value }.compact
  end   

  private

  def self.extract_info(more_info, required_keys, transform: nil)
    data = case more_info
           when Array
             more_info.first
           when String
             begin
               parsed = JSON.parse(more_info, symbolize_names: true)
               parsed.is_a?(Array) ? parsed.first : parsed
             rescue JSON::ParserError
               nil
             end
           when Hash
             more_info
           else
             nil
           end

    return nil unless data && required_keys.all? { |key| data.key?(key) }
    transform ? transform.call(data) : data
  end

  def self.extract_event_info(more_info)
    extract_info(more_info, [:source, :target, :sensor])
  end

  def self.extract_flow_info(more_info)
    extract_info(more_info, [:lan_ip, :wan_ip, :sensor])
  end

  def self.extract_vault_info(more_info)
    extract_info(more_info, [:source, :target, :sensor])
  end

  def self.extract_ap_state_info(more_info)
    extract_info(more_info, [:wireless_station, :sensor])
  end

  def self.extract_monitor_info(more_info)
    result = extract_info(more_info, [:value, :sum]) do |data|
      {
        value:       data[:value],
        sum:         data[:sum],
        min:         data[:min],
        max:         data[:max],
        monitor_sum: data[:monitor_sum],
        sensor_name: data[:sensor_name],
        timestamp:   data[:timestamp]
      }
    end

    Rails.logger.warn "Datos inválidos en extract_monitor_info: #{more_info.inspect}" unless result
    result
  end
  
  def self.dimensions_for(product_type)
    case product_type
    when 'flow'
      %w(lan_ip wan_ip sensor_name)
    when 'ips'
      %w(src dst sensor_name msg)
    when 'monitor'
      %w(sensor_name)
    when 'vault'
      %w(sensor_name source target hostname)
    when 'ap_state', 'wireless'
      %w(sensor_name wireless_station)
    else
      data_source::COLUMNS.keys rescue []
    end
  end
end
