#######################################################################
# Copyright (c) 2014 ENEO Tecnologia S.L.
# This file is part of redBorder.
# redBorder is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# redBorder is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
# You should have received a copy of the GNU Affero General Public License
# along with redBorder. If not, see <http://www.gnu.org/licenses/>.
#######################################################################

class IncidentGroupReport < ApplicationRecord
  include Exportable

  # ==========================================================
  # Custom attributes
  # ==========================================================

  attr_accessor :collaborators
  attr_accessor :sensors_collaborators
  exportable_opts relations: [:widgets]

  # ==========================================================
  # Relations
  # ==========================================================

  # Has many sensors and users
  has_many :report_users, dependent: :destroy
  has_many :report_sensors, dependent: :destroy
  has_many :users, through: :report_users
  has_many :sensors, through: :report_sensors

  # Created by an user
  # Alias
  belongs_to :owner, class_name: 'User', foreign_key: :user_id
  
  belongs_to :user

  # Has many report historicals.
  has_many :incident_reports_historicals, dependent: :destroy
  before_destroy :delete_historicals

  # We receive the days like a serialized array
  serialize :schedule_days

  # ================================================================
  # Callbacks
  # ================================================================

  before_save :save_collaborators
  before_save :save_sensors_collaborators


  # ================================================================
  # Validations
  # ================================================================

  # The name can't be empty
  validates_presence_of :name

  # Only allow one report per user on community version

  # ================================================================
  # URL helpers
  # ================================================================

  delegate :url_helpers, to: 'Rails.application.routes'

  # ================================================================
  # Methods
  # ================================================================

  def pages
    (1..number_pages).each_with_object([]) do |page, memo|
      memo << widgets.select { |w| w.page == page }
    end
  end

  def each_page
    (1..number_pages).each do |page|
      yield page
    end
  end


  #
  # Check if the user has permissions on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def permission?(user = self.user)
    return false if user.nil?
    user.writable_reports.include? self
  end

  #
  # Check if the user has read permission on this report.
  #
  # == Paramenters:
  # user:
  #   User who wants to edit a report
  #
  # == Returns:
  # A boolean
  #
  def read_permission?(user = self.user)
    return false if user.nil?
    user.readable_reports.include? self
  end

  #
  # Main URL to redirect when the permissions are wrong
  #
  # == Returns:
  # An url to redirect
  #
  def main_path
    url_helpers.reports_path
  end

  # ================================================================
  # Import/export methods for exportable concern
  # ================================================================

  #
  # => The attributes of this object that will be exported
  #
  def export_json
    to_json(except: [
      :id, :user_id, :created_at, :updated_at, :image_content_type,
      :image_file_size, :image_updated_at, :content_type, :time
    ])
  end

  def create_report_historical(username, file)
    # reports_historical = ReportsHistorical.new
    # reports_historical.report_id = id
    # reports_historical.generated_by = username
    # reports_historical.pdf_file = file
    # if reports_historical.save
    #   Rails.logger.info "Report: '#{id}' generated."
    # else
    #   Rails.logger.error "Report '#{id}' couldn't be generated."
    # end
  end

  # ================================================================
  # Private methods
  # ================================================================

  private

  def save_collaborators
    # Remove all relations, except those with the owner and the current user
    report_users.destroy_all
    return if collaborators.nil?

    # Dont save relations with the owner or the current user
    collaborators.each do |user|
      report_users.build(
        user_id: user['user_id'],
        can_edit: (user['can_edit'] == 'on')
      )
    end
  end

  def save_sensors_collaborators
    # Remove all relations
    report_sensors.destroy_all
    return if sensors_collaborators.nil?

    # Save all relations
    sensors_collaborators.each do |_, sensor|
      report_sensors.build(
        sensor_id: sensor['sensor_id'],
        can_edit: sensor['can_edit']
      )
    end
  end

  def delete_historicals
    Rails.logger.error "\n\nDeleting historicals\n\n"
    incident_reports_historicals.destroy_all
  end
end